@@ -43,6 +43,10 @@ typedef enum {
 	CD_MEDIA_TYPE_DVD_PLUS_RW,
 } CDMediaType;
 
+#define CD_MEDIA_SIZE_UNKNOWN -1
+#define CD_MEDIA_SIZE_NA      -2
+#define CD_MEDIA_SIZE_BUSY    -3
+
 typedef enum {
 	CDDRIVE_TYPE_FILE                   = 1 << 0,
 	CDDRIVE_TYPE_CD_RECORDER            = 1 << 1,
@@ -68,11 +72,15 @@ typedef struct {
 	CDDrivePriv *priv;
 } CDDrive;
 
+#define SIZE_TO_TIME(size) (int) (((size / 1024 / 1024) - 1) * 48 / 7)
+
 /* Returns a list of CDDrive structs */
 GList *scan_for_cdroms (gboolean recorder_only, gboolean add_image);
 void cd_drive_free (CDDrive *drive);
 CDMediaType cd_drive_get_media_type (CDDrive *drive);
 CDMediaType cd_drive_get_media_type_from_path (const char *device_path);
+gint64 cd_drive_get_media_size (CDDrive *drive);
+gint64 cd_drive_get_media_size_from_path (const char *device_path);
 CDDrive *cd_drive_get_file_image (void);
 
 #endif
