@@ -21,22 +21,12 @@
 #endif /* __FreeBSD__ */
 #include <signal.h>
 
+#ifdef USE_HAL
+#include <libhal.h>
+#endif
+
 #include "cd-recorder.h"
-
-typedef enum {
-	PREPARING_TO_WRITE_CD,
-	WRITING_CD,
-	FIXATING_CD,
-	BLANKING_CD,
-} ActionType;
-
-static char *actions [] = {
-	N_("Preparing to write CD"),
-	N_("Writing CD"),
-	N_("Fixating CD"),
-	N_("Erasing CD"),
-	NULL
-};
+#include "cd-recorder-marshal.h"
 
 struct CDRecorderPrivate {
 	GMainLoop *loop;
@@ -46,16 +36,19 @@ struct CDRecorderPrivate {
 	GString *line;
 	GString *cdr_stderr;
 	gboolean changed_text;
-	gboolean send_return;
 	gboolean expect_cdrecord_to_die;
-	GList *actions;
 	gboolean dangerous;
 
 	char *last_error;
-	
+
 	GList *tracks;
 	int track_count;
 	gboolean debug;
+	gboolean can_rewrite;
+
+#ifdef USE_HAL
+	LibHalContext *ctx;
+#endif
 };
 
 /* Signals */
@@ -64,7 +57,6 @@ enum {
 	ACTION_CHANGED,
 	ANIMATION_CHANGED,
 	INSERT_CD_REQUEST,
-	INSERT_RWCD_REQUEST,
 	LAST_SIGNAL
 };
 
@@ -98,6 +90,8 @@ cd_recorder_track_free (Track *track)
 	default:
 		g_warning ("Invalid track type %d", track->type);
 	}
+
+	g_free (track);
 }
 
 static gboolean
@@ -113,9 +107,8 @@ can_burn_dvds (CDDrive *recorder)
 }
 
 static gboolean
-cd_needs_growisofs (CDDrive *recorder, GList *tracks)
+cd_needs_growisofs (CDDrive *recorder, CDMediaType type, GList *tracks)
 {
-	CDMediaType type;
 	GList *l;
 
 	/* If we cannot burn DVDs, then we don't need growisofs */
@@ -123,7 +116,6 @@ cd_needs_growisofs (CDDrive *recorder, GList *tracks)
 		return FALSE;
 	}
 
-	type = cd_drive_get_media_type (recorder);
 	if (type == CD_MEDIA_TYPE_DVDR
 			|| type == CD_MEDIA_TYPE_DVDRW
 			|| type == CD_MEDIA_TYPE_DVD_PLUS_R
@@ -157,21 +149,6 @@ cd_recorder_cancel (CDRecorder *cdrecorder, gboolean skip_if_dangerous)
 	return cdrecorder->priv->dangerous;
 }
 
-void
-cd_recorder_insert_cd_retry (CDRecorder *cdrecorder, gboolean is_reload)
-{
-	if (is_reload) {
-		if (cdrecorder->priv->send_return) {
-			write (cdrecorder->priv->cdr_stdin, "\n", 1);
-		} else {
-			kill (cdrecorder->priv->pid, SIGUSR1);
-		}
-	} else {
-		cdrecorder->priv->result = RESULT_RETRY;
-		g_main_loop_quit (cdrecorder->priv->loop);
-	}
-}
-
 static gboolean
 growisofs_stdout_read (GIOChannel   *source,
 		      GIOCondition  condition,
@@ -205,7 +182,7 @@ growisofs_stdout_read (GIOChannel   *source,
 			if (!cdrecorder->priv->changed_text) {
 				g_signal_emit (G_OBJECT (cdrecorder),
 					       cd_recorder_table_signals[ACTION_CHANGED], 0,
-					       WRITING_CD);
+					       WRITING, MEDIA_DVD);
 			}
 
 			percent = (perc_1 + ((float) perc_2 / 10)) / 100;
@@ -266,7 +243,7 @@ cdrecord_stdout_read (GIOChannel   *source,
 			if (!cdrecorder->priv->changed_text) {
 				g_signal_emit (G_OBJECT (cdrecorder),
 					       cd_recorder_table_signals[ACTION_CHANGED], 0,
-					       WRITING_CD);
+					       WRITING, MEDIA_CD);
 			}
 			percent = 0.98 * ((double)((track-1)/(double)cdrecorder->priv->track_count)
 					  + ((double)mb_written/mb_total) / (double)cdrecorder->priv->track_count);
@@ -275,14 +252,11 @@ cdrecord_stdout_read (GIOChannel   *source,
 				       percent);
 		} else if (g_str_has_prefix (line, "Re-load disk and hit <CR>") ||
 			   g_str_has_prefix (line, "send SIGUSR1 to continue")) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
-				       TRUE);
-			cdrecorder->priv->send_return = (*line == 'R');
+			g_assert_not_reached ();
 		} else if (g_str_has_prefix (line, "Fixating...")) {
 			g_signal_emit (G_OBJECT (cdrecorder),
 				       cd_recorder_table_signals[ACTION_CHANGED], 0,
-				       FIXATING_CD);
+				       FIXATING, MEDIA_CD);
 		} else if (g_str_has_prefix (line, "Fixating time:")) {
 			g_signal_emit (G_OBJECT (cdrecorder),
 				       cd_recorder_table_signals[PROGRESS_CHANGED], 0,
@@ -293,7 +267,7 @@ cdrecord_stdout_read (GIOChannel   *source,
 		} else if (g_str_has_prefix (line, "Blanking PMA, TOC, pregap")) {
 			g_signal_emit (G_OBJECT (cdrecorder),
 				       cd_recorder_table_signals[ACTION_CHANGED], 0,
-				       BLANKING_CD);
+				       BLANKING, MEDIA_CD);
 		}
 
 		g_free (line);
@@ -333,10 +307,7 @@ growisofs_stderr_read (GIOChannel   *source,
 	if (status == G_IO_STATUS_NORMAL && !cdrecorder->priv->expect_cdrecord_to_die) {
 		g_string_prepend (cdrecorder->priv->cdr_stderr, line);
 		if (strstr (line, "unsupported MMC profile") != NULL || (strstr (line, "already carries isofs") != NULL && strstr (line, "FATAL:") != NULL)) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
-				       FALSE);
-			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
+			g_assert_not_reached ();
 		} else if (strstr (line, "unable to open") != NULL || strstr (line, "unable to stat") != NULL) {
 			/* This fits the "open64" and "open"-like messages */
 			cdrecorder->priv->last_error = g_strdup (_("The recorder could not be accessed"));
@@ -344,13 +315,16 @@ growisofs_stderr_read (GIOChannel   *source,
 		} else if (strstr (line, "not enough space available") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("Not enough space available on the disc"));
 			cdrecorder->priv->result = RESULT_ERROR;
+		} else if (strstr (line, "end of user area encountered on this track") != NULL) {
+			cdrecorder->priv->last_error = g_strdup (_("The files selected did not fit on the CD"));
+			cdrecorder->priv->result = RESULT_ERROR;
 		} else if (strstr (line, "blocks are free") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("The files selected did not fit on the CD"));
 			cdrecorder->priv->result = RESULT_ERROR;
 		} else if (strstr (line, "flushing cache") != NULL) {
 			g_signal_emit (G_OBJECT (cdrecorder),
 				       cd_recorder_table_signals[ACTION_CHANGED], 0,
-				       FIXATING_CD);
+				       FIXATING, MEDIA_DVD);
 			cdrecorder->priv->result = RESULT_FINISHED;
 		} else if (strstr (line, ":-(") != NULL || strstr (line, "FATAL") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("Unhandled error, aborting"));
@@ -362,7 +336,7 @@ growisofs_stderr_read (GIOChannel   *source,
 		}
 		return FALSE;
 	} else {
-		g_print ("cdrecord stderr read failed, status: %d\n", status);
+		g_print ("growisofs stderr read failed, status: %d\n", status);
 	}
 
 	return TRUE;
@@ -387,18 +361,14 @@ cdrecord_stderr_read (GIOChannel   *source,
 	if (status == G_IO_STATUS_NORMAL && !cdrecorder->priv->expect_cdrecord_to_die) {
 		g_string_prepend (cdrecorder->priv->cdr_stderr, line);
 		if (strstr (line, "No disk / Wrong disk!") != NULL) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
-				       FALSE);
-			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
+			g_warning ("No disk in drive, shouldn't happen");
+			g_assert_not_reached ();
 		} else if (strstr (line, "This means that we are checking recorded media.") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("The CD has already been recorded"));
 			cdrecorder->priv->result = RESULT_ERROR;
 		} else if (strstr (line, "Cannot blank disk, aborting.") != NULL) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_RWCD_REQUEST], 0,
-				       FALSE);
-			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
+			g_warning ("Right CD type wasn't loaded, shouldn't happen");
+			g_assert_not_reached ();
 		} else if (strstr (line, "Data may not fit on current disk") != NULL) {
 			cdrecorder->priv->last_error = g_strdup (_("The files selected did not fit on the CD"));
 			/* FIXME should we error out in that case?
@@ -407,10 +377,11 @@ cdrecord_stderr_read (GIOChannel   *source,
 			cdrecorder->priv->last_error = g_strdup (_("All audio files must be stereo, 16-bit digital audio with 44100Hz samples"));
 			cdrecorder->priv->result = RESULT_ERROR;
 		} else if (strstr (line, "cannot write medium - incompatible format") != NULL) {
-			g_signal_emit (G_OBJECT (cdrecorder),
-				       cd_recorder_table_signals[INSERT_CD_REQUEST], 0,
-				       FALSE);
+			g_warning ("Right CD type wasn't loaded, shouldn't happen");
+			g_assert_not_reached ();
 			cdrecorder->priv->expect_cdrecord_to_die = TRUE;
+		} else if (strstr (line, "DMA speed too slow") != NULL) {
+			cdrecorder->priv->last_error = g_strdup (_("The system is too slow to write the CD at this speed. Try a lower speed."));
 		}
 
 		g_free (line);
@@ -426,26 +397,133 @@ cdrecord_stderr_read (GIOChannel   *source,
 	return TRUE;
 }
 
+static gboolean
+media_type_matches (CDMediaType type, gboolean *needs_blank)
+{
+	*needs_blank = FALSE;
+
+	switch (type)
+	{
+	case CD_MEDIA_TYPE_ERROR:
+	case CD_MEDIA_TYPE_BUSY:
+		return FALSE;
+	case CD_MEDIA_TYPE_UNKNOWN:
+		return TRUE;
+	case CD_MEDIA_TYPE_CD:
+	case CD_MEDIA_TYPE_DVD:
+	case CD_MEDIA_TYPE_DVD_RAM:
+		return FALSE;
+	case CD_MEDIA_TYPE_CDR:
+		return TRUE;
+	case CD_MEDIA_TYPE_CDRW:
+		*needs_blank = TRUE;
+		return TRUE;
+	case CD_MEDIA_TYPE_DVDR:
+		return TRUE;
+	case CD_MEDIA_TYPE_DVDRW:
+		*needs_blank = TRUE;
+		return TRUE;
+	case CD_MEDIA_TYPE_DVD_PLUS_R:
+		return TRUE;
+	case CD_MEDIA_TYPE_DVD_PLUS_RW:
+		*needs_blank = TRUE;
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static CDMediaType
+cd_recorder_wait_for_insertion (CDRecorder *cdrecorder, CDDrive *drive,
+		gboolean *needs_blank)
+{
+	CDMediaType type;
+	gboolean reload;
+
+	reload = FALSE;
+	type = cd_drive_get_media_type (drive);
+
+	if (type == CD_MEDIA_TYPE_ERROR) {
+		reload = TRUE;
+	}
+
+	while (!media_type_matches (type, needs_blank)) {
+		gboolean res, busy_cd;
+
+		busy_cd = (type == CD_MEDIA_TYPE_BUSY);
+
+#ifdef USE_HAL
+		if (busy_cd != FALSE) {
+			/* TODO umount the disc if we find it,
+			 * and try it again */
+		}
+#endif
+
+		g_signal_emit (G_OBJECT (cdrecorder),
+				cd_recorder_table_signals[INSERT_CD_REQUEST],
+				0, reload, cdrecorder->priv->can_rewrite,
+				busy_cd, &res);
+
+		if (res == FALSE) {
+			return CD_MEDIA_TYPE_ERROR;
+		}
+
+		type = cd_drive_get_media_type (drive);
+		reload = FALSE;
+		if (type == CD_MEDIA_TYPE_UNKNOWN
+				|| type == CD_MEDIA_TYPE_ERROR) {
+			reload = TRUE;
+		}
+	}
+
+	return type;
+}
+
 int
 cd_recorder_write_tracks (CDRecorder *cdrecorder,
-		       CDDrive *recorder,
+		       CDDrive *drive,
 		       GList *tracks,
 		       int speed,
 		       CDRecorderWriteFlags flags)
 		       /* TODO: GError **error */
 {
+	CDMediaType type;
+	gboolean needs_blank;
+
 	g_return_val_if_fail (tracks != NULL, RESULT_ERROR);
 	cdrecorder->priv->tracks = tracks;
 	cdrecorder->priv->track_count = g_list_length (tracks);
 	cdrecorder->priv->debug = (flags & CDRECORDER_DEBUG);
 
-	if (cd_needs_growisofs (recorder, tracks)) {
+	/* Let's hope that DVD+RW and DVD-RW drives can also do CD-RWs */
+	cdrecorder->priv->can_rewrite =
+		(drive->type & CDDRIVE_TYPE_CDRW_RECORDER);
+
+	if (cdrecorder->priv->track_count > 99) {
+		cdrecorder->priv->last_error = g_strdup (_("You can only burn 99 tracks on one disc"));
+		cdrecorder->priv->result = RESULT_ERROR;
+		return cdrecorder->priv->result;
+	}
+
+	type = cd_recorder_wait_for_insertion (cdrecorder,
+			drive, &needs_blank);
+
+	if (type == CD_MEDIA_TYPE_ERROR) {
+		cdrecorder->priv->result = RESULT_CANCEL;
+		return cdrecorder->priv->result;
+	}
+
+	if (needs_blank != FALSE) {
+		flags |= CDRECORDER_BLANK;
+	}
+
+	if (cd_needs_growisofs (drive, type, tracks)) {
 		return cd_recorder_write_growisofs (cdrecorder,
-						    recorder, tracks, speed,
+						    drive, tracks, speed,
 						    flags);
 	} else {
 		return cd_recorder_write_cdrecord (cdrecorder,
-						   recorder, tracks, speed,
+						   drive, tracks, speed,
 						   flags);
 	}
 }
@@ -511,7 +589,7 @@ retry_growisofs:
 
 	g_signal_emit (G_OBJECT (cdrecorder),
 			cd_recorder_table_signals[ACTION_CHANGED], 0,
-			PREPARING_TO_WRITE_CD);
+			PREPARING_WRITE, MEDIA_DVD);
 	cdrecorder->priv->changed_text = FALSE;
 	g_signal_emit (G_OBJECT (cdrecorder),
 			cd_recorder_table_signals[PROGRESS_CHANGED], 0,
@@ -525,7 +603,7 @@ retry_growisofs:
 	{
 		guint i;
 		g_print ("launching command: ");
-		for (i = 0; i < argv->len; i++) {
+		for (i = 0; i < argv->len - 1; i++) {
 			g_print ("%s ", (char*)g_ptr_array_index (argv, i));
 		}
 		g_print ("\n");
@@ -544,7 +622,7 @@ retry_growisofs:
 					&error)) {
 		guint i;
 		g_warning ("growisofs command failed: %s\n", error->message);
-		for (i = 0; i < argv->len; i++) {
+		for (i = 0; i < argv->len - 1; i++) {
 			g_print ("%s ", (char*)g_ptr_array_index (argv, i));
 		}
 		g_print ("\n");
@@ -642,6 +720,12 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 	if (flags & CDRECORDER_DISC_AT_ONCE) {
 		g_ptr_array_add (argv, "-dao");
 	}
+	if (flags & CDRECORDER_OVERBURN) {
+		g_ptr_array_add (argv, "-overburn");
+	}
+	if (flags & CDRECORDER_BURNPROOF) {
+		g_ptr_array_add (argv, "driveropts=burnfree");
+	}
 	g_ptr_array_add (argv, "-v");
 
 	l = tracks;
@@ -680,7 +764,7 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 
 	g_signal_emit (G_OBJECT (cdrecorder),
 		       cd_recorder_table_signals[ACTION_CHANGED], 0,
-		       PREPARING_TO_WRITE_CD);
+		       PREPARING_WRITE, MEDIA_CD);
 	cdrecorder->priv->changed_text = FALSE;
 	g_signal_emit (G_OBJECT (cdrecorder),
 		       cd_recorder_table_signals[PROGRESS_CHANGED], 0,
@@ -694,7 +778,7 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 	{
 		guint i;
 		g_print ("launching command: ");
-		for (i = 0; i < argv->len; i++) {
+		for (i = 0; i < argv->len - 1; i++) {
 			g_print ("%s ", (char*)g_ptr_array_index (argv, i));
 		}
 		g_print ("\n");
@@ -760,12 +844,6 @@ cd_recorder_write_cdrecord (CDRecorder *cdrecorder,
 	return cdrecorder->priv->result;
 }
 
-GList *
-cd_recorder_list_actions (CDRecorder *cdrecorder)
-{
-	return cdrecorder->priv->actions;
-}
-
 const char *
 cd_recorder_get_error_message_details (CDRecorder *cdrecorder)
 {
@@ -789,8 +867,15 @@ cd_recorder_finalize (GObject *object)
 
 	g_return_if_fail (object != NULL);
 
-	g_string_free (cdrecorder->priv->cdr_stderr, TRUE);
-	/* FIXME cleanup */
+	if (cdrecorder->priv->cdr_stderr != NULL) {
+		g_string_free (cdrecorder->priv->cdr_stderr, TRUE);
+	}
+
+	if (cdrecorder->priv->line != NULL) {
+		g_string_free (cdrecorder->priv->line, TRUE);
+	}
+
+	g_free (cdrecorder->priv->last_error);
 
 	if (G_OBJECT_CLASS (parent_class)->finalize != NULL) {
 		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
@@ -800,16 +885,7 @@ cd_recorder_finalize (GObject *object)
 static void
 cd_recorder_init (CDRecorder *cdrecorder)
 {
-	int i;
-
 	cdrecorder->priv = g_new0 (CDRecorderPrivate, 1);
-
-	for (i = 0; actions[i] != NULL; i++)
-	{
-		cdrecorder->priv->actions = g_list_append
-			(cdrecorder->priv->actions,
-			 _(actions[i]));
-	}
 }
 
 CDRecorder *
@@ -842,8 +918,8 @@ cd_recorder_class_init (CDRecorderClass *klass)
 				G_STRUCT_OFFSET (CDRecorderClass,
 					action_changed),
 				NULL, NULL,
-				g_cclosure_marshal_VOID__INT,
-				G_TYPE_NONE, 1, G_TYPE_INT);
+				cdrecorder_marshal_VOID__INT_INT,
+				G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);
 	cd_recorder_table_signals[ANIMATION_CHANGED] =
 		g_signal_new ("animation-changed",
 				G_TYPE_FROM_CLASS (klass),
@@ -860,17 +936,8 @@ cd_recorder_class_init (CDRecorderClass *klass)
 				G_STRUCT_OFFSET (CDRecorderClass,
 					insert_cd_request),
 				NULL, NULL,
-				g_cclosure_marshal_VOID__BOOLEAN,
-				G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
-	cd_recorder_table_signals[INSERT_RWCD_REQUEST] =
-		g_signal_new ("insert-rwcd-request",
-				G_TYPE_FROM_CLASS (klass),
-				G_SIGNAL_RUN_LAST,
-				G_STRUCT_OFFSET (CDRecorderClass,
-					insert_cd_rw_request),
-				NULL, NULL,
-				g_cclosure_marshal_VOID__BOOLEAN,
-				G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
-
+				cdrecorder_marshal_BOOLEAN__BOOLEAN_BOOLEAN_BOOLEAN,
+				G_TYPE_BOOLEAN, 3, G_TYPE_BOOLEAN,
+				G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
 }
 
